#include "pch.h"
#include "common.h"
#include "TexturePackerSupport.h"

DevTexture g_TexturePackerTexture;

TexturePackerDocument::TexturePackerDocument()
{
}

TexturePackerDocument::~TexturePackerDocument()
{
}

void TexturePackerDocument::Load(const char* filePath)
{
    std::vector<byte> fileBytes;
    bool ok = base::NFS.GetFileBytes(filePath, fileBytes);
    assert(ok);

    rapidjson::Document doc;
    doc.Parse((const char*)fileBytes.data(), fileBytes.size());
    assert(!doc.HasParseError());
    assert(doc.IsObject());

    auto metaVal = doc.FindMember("meta");
    assert(metaVal != doc.MemberEnd() && metaVal->value.IsObject());
    LoadMeta(metaVal->value);
    
    auto framesVal = doc.FindMember("frames");
    assert(framesVal != doc.MemberEnd() && framesVal->value.IsObject());
    for(auto frameIt = framesVal->value.MemberBegin();
        frameIt != framesVal->value.MemberEnd(); ++frameIt)
    {
        assert(frameIt->name.IsString());
        LoadFrame(frameIt->name.GetString(), frameIt->value);
    }
}

size_t TexturePackerDocument::FrameFileNameToIndex(const char* fileName) const
{
    for(size_t i = 0, count = m_Frames.size(); i < count; ++i)
    {
        if(_stricmp(fileName, m_Frames[i].fileName.c_str()) == 0)
        {
            return i;
        }
    }
    return SIZE_MAX;
}

void TexturePackerDocument::GetTexCoords(float* outTexCoords, size_t frameIndex) const
{
    const Frame& frame = m_Frames[frameIndex];
    outTexCoords[0] = (float)frame.frame.x * m_SizeX_inv;
    outTexCoords[1] = (float)frame.frame.y * m_SizeY_inv;
    outTexCoords[2] = outTexCoords[0] + (float)frame.frame.w * m_SizeY_inv;
    outTexCoords[3] = outTexCoords[1] + (float)frame.frame.h * m_SizeY_inv;
}

void TexturePackerDocument::LoadMeta(const rapidjson::Value& val)
{
    LoadParam_Size(m_Size, val, "size");

    m_SizeX_inv = 1.f / (float)m_Size.w;
    m_SizeY_inv = 1.f / (float)m_Size.h;
}

void TexturePackerDocument::LoadFrame(const char* fileName, const rapidjson::Value& val)
{
    Frame newFrame;
    newFrame.fileName = fileName;

    assert(val.IsObject());
    LoadParam_Rect(newFrame.frame, val, "frame");

    bool b;
    LoadParam_Bool(b, val, "rotated");
    assert(!b);
    LoadParam_Bool(b, val, "trimmed");
    assert(!b);

    Rect spriteSourceSize;
    LoadParam_Rect(spriteSourceSize, val, "spriteSourceSize");
    assert(spriteSourceSize.x == 0 && spriteSourceSize.y == 0 &&
        spriteSourceSize.w == newFrame.frame.w &&
        spriteSourceSize.h == newFrame.frame.h);

    Size sourceSize;
    LoadParam_Size(sourceSize, val, "sourceSize");
    assert(sourceSize.w == newFrame.frame.w &&
        sourceSize.h == newFrame.frame.h);

    m_Frames.push_back(std::move(newFrame));
}

void TexturePackerDocument::LoadParam_Bool(bool& out, const rapidjson::Value& objVal, const char* paramName)
{
    auto paramIt = objVal.FindMember(paramName);
    assert(paramIt != objVal.MemberEnd());

    assert(paramIt->value.IsBool());
    out = paramIt->value.GetBool();
}

void TexturePackerDocument::LoadParam_Int(int& out, const rapidjson::Value& objVal, const char* paramName)
{
    auto paramIt = objVal.FindMember(paramName);
    assert(paramIt != objVal.MemberEnd());

    assert(paramIt->value.IsInt());
    out = paramIt->value.GetInt();
}

void TexturePackerDocument::LoadParam_Rect(Rect& out, const rapidjson::Value& objVal, const char* paramName)
{
    auto paramIt = objVal.FindMember(paramName);
    assert(paramIt != objVal.MemberEnd());

    assert(paramIt->value.IsObject());
    LoadParam_Int(out.x, paramIt->value, "x");
    LoadParam_Int(out.y, paramIt->value, "y");
    LoadParam_Int(out.w, paramIt->value, "w");
    LoadParam_Int(out.h, paramIt->value, "h");
}

void TexturePackerDocument::LoadParam_Size(Size& out, const rapidjson::Value& objVal, const char* paramName)
{
    auto paramIt = objVal.FindMember(paramName);
    assert(paramIt != objVal.MemberEnd());

    assert(paramIt->value.IsObject());
    LoadParam_Int(out.w, paramIt->value, "w");
    LoadParam_Int(out.h, paramIt->value, "h");
}

void Animation::Set(
    const char* const* frameNames,
    size_t frameCount,
    float singleFrameTime)
{
    m_SingleFrameTime = singleFrameTime;
    m_TotalTime = singleFrameTime * (float)frameCount;

    for(size_t i = 0; i < frameCount; ++i)
    {
        const size_t frameIndex = game.m_TexturePackerDoc.FrameFileNameToIndex(frameNames[i]);
        assert(frameIndex != SIZE_MAX);
        m_FrameIndices.push_back(frameIndex);
    }
}

size_t Animation::GetCurrentFrameIndex(float t) const
{
    const float wrappedT = fmodf(t, m_TotalTime);
    const size_t indexIndex = min((size_t)(wrappedT / m_SingleFrameTime), m_FrameIndices.size());
    return m_FrameIndices[indexIndex];
}
