﻿#include "pch.h"
#include "common.h"
#include "InputManager.h"
#include "FmodSound.hpp"
#include <thread>


DevRenderTarget main_rt(D3DFMT_A8R8G8B8,0,0,1);
DevRenderTarget heat_rt(D3DFMT_A8R8G8B8,0,0,1);


DevEffect fx("traceball.fx",NULL,FXF_FLOW_CONTROL_MEDIUM|FXF_OPTIMIZATION_1);
DevTexture t_random;


Engine game;

static const char* const TEXTURE_ATLAS_JSON_FILE_PATH = "data/lazerbugz_hash.json";
const char* const TEXTURE_ATLAS_TEXTURE_FILE_PATH = "data/lazerbugz_hash.png";



const char *editor_help_text[] = {
	"",
	"Editor keys:",
	"  LPM - Draw common ground",
	"  RPM - Draw high ground",
	"  MPM - Draw lava",
	"  1-5, tilde - Brush sizes (with shift a little larger)",
	"  B - Place ball 1",
	"  N - Place ball 2",
	"  T - Place exit 1",
	"  Y - Place exit 2",
	"  G - Change ball 1 type",
	"  H - Change ball 2 type",
	"  J - Change high ground material",
	"  K - Change common ground material",
	"  L - Change lava material",
	"  C - Set starting camera position",
	"  X - Move camera to starting position",
	"  unused A-Z - Drawing macros",
	"  F1 - Help on/off",
	"  F4 - Editor mode on/off",
	"  F5 - Load",
	"  F6 - Save",
	"  F7 - Previous level",
	"  F8 - Next level",
	"  Alt+F4 - Exit game",
	NULL
};

bool editor_help = false;
float g_time;
int g_maxlevel;
int g_skiplevel;



Engine::Engine()
{
    editor_mode = false;
    exit_status=vec3(0, 0, 0);
    m_ExtendedShaderFeaturesEnabled = GetConfigInt("Graphics", "ExtendedShaderFeatures", "1");
}

void Engine::Init()
{
    m_TexturePackerDoc.Load(TEXTURE_ATLAS_JSON_FILE_PATH);
    InitAnimations();
    InitSoundEffects();
}

void Engine::InitLevel(int level,int mode)
{
	level_id = level;
	completed = 0;
	exit_status = vec3(0,0,0);

	if(mode & INIT_CONFIG)
	{
		messages.clear();
		raw_cmd.clear();
		ProcessConfig("maps/default.cfg",mode);
		ProcessConfig(format("maps/%04d.cfg",level).c_str(),mode);
	}

	ball[0].Init(locations[0]);
	ball[1].Init(locations[1]);
	ball[2].Init(locations[2]);
	ball[3].Init(locations[3]);
	ball[4].Init(locations[4]);

	{
		int cam_i = cameras.size();
		if( mode & INIT_CAMERA ) cam_i = 0;
		
		cameras.resize(editor_mode ? 1 : 1);
		for( ;cam_i<(int)cameras.size(); cam_i++ )
		{
			cameras[cam_i].dir_front = locations[2].get_normalized();
			cameras[cam_i].dir_up = locations[3].get_normalized();
			cameras[cam_i].active = (cam_i==0);
			cameras[cam_i].attach_obj = NULL;
		}
	}

	if(mode & INIT_MAP)
		chmap.Load(format("maps/%04d.bin", level).c_str());
	
	if( !editor_mode )
		sq.Run("on_map_init");

	if(mode & INIT_MATERIALS)
	{
		mat[0].Load(mat[0].mname.c_str());
		mat[1].Load(mat[1].mname.c_str());
		mat[2].Load(mat[2].mname.c_str());
	}

	if((mode & INIT_RESET_COMPLETED) && completed)
	{
		completed = 0;
		WriteConfig(format("maps/%04d.cfg",level_id).c_str());
	}

	if(mode & INIT_MESSAGES)
		if(completed)
		{
			if(messages.size()>=2)
				ui.PopMessage(messages[messages.size()-1].text.c_str(),messages[messages.size()-1].width);
		}
		else
			for(int i=0;i+2<(int)messages.size();i++)
				ui.PopMessage(messages[i].text.c_str(),messages[i].width);

	if(completed)
		exit_status = vec3(10,10,1);
}

void Engine::ProcessConfig(const char *path,int mode)
{
	vector<string> cfg;

	if(!NFS.GetFileLines(path,cfg))
		return;

	static char scmd[100],b2[1000];
	int id;
	for(int i=0;i<(int)cfg.size();i++)
	{
		const char *line = cfg[i].c_str();
		scmd[0] = 0;
		id = 0;
		sscanf(line,"%s%d",scmd,&id);
		string cmd = scmd;

		b2[0] = 0;

		if(cmd=="loc" && id>=0 && id<MAX_LOCATIONS)
			sscanf(line,"%s%d%f%f%f",scmd,&id,&locations[id].x,&locations[id].y,&locations[id].z);

		if(cmd=="mat" && id>=0 && id<=2)
		{
			sscanf(line,"%s%d%s",scmd,&id,b2);
			mat[id].mname = b2;
		}

		if(cmd=="balls")
			sscanf(line,"%s%d%d",scmd,&ball_types[0],&ball_types[1]);

		if(cmd=="mdclear")
			material_list.clear();

		if(cmd=="mdef")
		{
			sscanf(line,"%s%s",scmd,b2);
			material_list.push_back(b2);
		}

		if(cmd=="msg")
		{
			const char *s = line;
			while(*s && *s!=' ') s++;
			while(*s==' ') s++;
			while(*s && *s!=' ') s++;
			while(*s==' ') s++;
			
			GameUI::Message m;
			m.text = s;
			m.width = id;
			messages.push_back(m);

			raw_cmd.push_back(line);
		}

		if(cmd=="+" && messages.size()>0)
		{
			messages.back().text.append(line+1);
			raw_cmd.push_back(line);
		}

		if(cmd=="maxlevel")
			g_maxlevel = id;

		if(cmd=="completed")
			completed = 1;

		if( cmd=="draw" && cameras.size()>0 )
		{
			float r=0.05f,h=0.85f,pwr=1,ofs=0,ang=0;
			sscanf(line,"%s%f%f%f%f%f",scmd,&r,&h,&pwr,&ofs,&ang);
			vec3 pos = -cameras[0].dir_front;
			pos += cameras[0].dir_up*(ofs*cos(ang*2*3.141593f));
			pos += cameras[0].dir_up.cross(cameras[0].dir_front)*(ofs*cos(ang*2*3.141593f));
			chmap.Draw(pos,r,h,pwr);
		}
	}
}

void Engine::WriteConfig(const char *path)
{
	FILE *fp = fopen(path,"wt");
	if(!fp) return;

	fprintf(fp,"# Autogenerated level file\n");
	
	for(int i=0;i<=2;i++)
		fprintf(fp,"mat %d %s\n",i,mat[i].mname.c_str());
	for(int i=0;i<MAX_LOCATIONS;i++)
		fprintf(fp,"loc %d %f %f %f\n",i,locations[i].x,locations[i].y,locations[i].z);
	fprintf(fp,"balls %d %d\n",ball_types[0],ball_types[1]);
	for(int i=0;i<(int)raw_cmd.size();i++)
		fprintf(fp,"%s\n",raw_cmd[i].c_str());
	if(completed)
		fprintf(fp,"completed\n");

	fclose(fp);
}


void Engine::Physics(float delta)
{
	if(editor_mode || completed)
		return;


	for( int i=0; i<(int)objects.size(); i++ )
	{
		EngineObject *ob = objects[i].get();
		if( !ob )
		{
			objects[i--] = objects.back();
			objects.pop_back();
			continue;
		}

		ob->TickPhysics(delta);
	}


//	ball[0].Physics(chmap,-dir_front,ball[1],delta);
//	ball[1].Physics(chmap,-dir_front,ball[0],delta);
//
//	bool cmp = (exit_status.z<1);
//#define EXIT(v,m)		if(v>0) { v+=delta*m*0.2f; if(v>m) v=m; }
//	EXIT(exit_status.x,10)
//	EXIT(exit_status.y,10)
//	EXIT(exit_status.z,1)
//#undef EXIT
//
//	if(cmp && exit_status.z>=1)
//	{
//		if(!completed)
//		{
//			completed = 1;
//			WriteConfig(format("maps/%04d.cfg",level_id).c_str());
//		}
//		if(messages.size()>=2)
//			ui.PopMessage(messages[messages.size()-2].text.c_str(),messages[messages.size()-2].width);
//	}
//
//
//	if((ball[1].pos-ball[0].pos).length()<0.05f*2)
//	{
//		vec3 p = (ball[0].pos + ball[1].pos)*0.5f;
//		vec3 n = (ball[1].pos - ball[0].pos).get_normalized();
//		ball[0].pos  = p - n*(0.05f*1.01f);
//		ball[1].pos = p + n*(0.05f*1.01f);
//
//		if(n.dot(ball[1].vel-ball[0].vel)<0)
//		{
//			float v1 = n.dot(ball[0].vel);
//			float v2 = n.dot(ball[1].vel);
//			ball[0].vel  += n*(v2-v1);
//			ball[1].vel += n*(v1-v2);
//			ball[0].Touch(ball[1]);
//			ball[1].Touch(ball[0]);
//		}
//
//		ball[0].Physics(chmap,-dir_front,ball[1],0.f);
//		ball[1].Physics(chmap,-dir_front,ball[0],0.f);
//	}
//
//	for(int i=0;i<2;i++)
//		if(ball[i].pos.length()<0.73f)
//			for(int l=0;l<2;l++)
//				if((&exit_status.x)[l]<=0.f && ball[i].pos.get_normalized().dot(locations[4+l])>=0.97f)
//				{
//					(&exit_status.x)[l] += 0.001f;
//					if(ball[i].type==BALL_ENEMY)
//						ball[i].type = BALL_DEAD;
//				}
//
//	if(exit_status.x>=2.5f && exit_status.y>=2.5f && exit_status.z<=0)
//		exit_status.z += 0.001f;
	
}

void EngineCamera::RunCamera(float delta)
{
	if( attach_obj.get() && !game.editor_mode )
	{
		dir_front = -attach_obj->pos;
	}
	else if( game.editor_mode )
	{
		int mx=0, my=0;
		Dev.GetMouseDelta(mx, my);	// Sio z tym!	[KK: OK!]

		mouse_smooth += vec2((float)mx, (float)my);
		vec2 dm = mouse_smooth*(1-pow(0.000000001f, delta));
		mouse_smooth -= dm;

		dm *= 0.01f/3;
		dir_front += dir_up.cross(dir_front)*dm.x;
		dir_front += dir_up*dm.y;
	}

	dir_front.normalize();
	dir_up -= dir_front * dir_front.dot(dir_up);
	dir_up.normalize();
}

void Engine::RunGame(float delta)
{
	for( int i=0; i<(int)cameras.size(); i++ )
		cameras[i].RunCamera(delta);

	int ch;
	while((ch=Dev.ReadKey()))
	{
		if(ch=='R') InitLevel(level_id,INIT_LEVEL_RESTART);
		if(ch==VK_F4) editor_mode = true, InitLevel(level_id,INIT_EDITOR_ENTER);
		if(ch==VK_F7 && level_id>0) InitLevel(level_id-1,INIT_ALL);
		if(ch==VK_F8 && level_id<g_maxlevel) InitLevel(level_id+1,INIT_ALL);
	}

	// Start frame
	sq.Run("on_frame");

	// Step physics
	{
		float step_delta = delta;
		int steps = (int)ceil(step_delta/0.1f);
		if( steps>0 ) step_delta /= steps;
		if( steps>10 ) steps = 10;
		for( int i=0; i<steps; i++ )
			game.Physics(step_delta);
	}

	// Update objects
	for( int i=0; i<(int)objects.size(); i++ )
	{
		EngineObject *ob = objects[i].get();
		if( !ob )
		{
			objects[i--] = objects.back();
			objects.pop_back();
			continue;
		}

		if( delta >= ob->update_time )
		{
			ob->update_time -= delta;
			if( ob->update_time < -0.1f )	// no more than 100ms overdue
				ob->update_time = -0.1f;

			Sqrat::RootTable(sq()).SetValue("self", ob);
			sq.Run("on_object_update");
		}
		else
			ob->update_time -= delta;
	}

	// End frame
	sq.Run("on_after_frame");
}

void Engine::RunEditor(float delta)
{
	for( int i=0; i<(int)cameras.size(); i++ )
		cameras[i].RunCamera(delta);

	static float radius = 0.010f;
	if( cameras.size()>0 )
	{
		if( Dev.GetKeyState(VK_LBUTTON) )	chmap.Draw(-cameras[0].dir_front, radius, 0.85f);
		if( Dev.GetKeyState(VK_RBUTTON) )	chmap.Draw(-cameras[0].dir_front, radius, 1.00f);
		if( Dev.GetKeyState(VK_MBUTTON) )	chmap.Draw(-cameras[0].dir_front, radius, 0.70f);
	}

	bool shift = Dev.GetKeyState(VK_LSHIFT);
	int ch;
	while((ch=Dev.ReadKey()))
	{
		if(ch==VK_F1) editor_help = !editor_help;
		if(ch==VK_F3) chmap.ExportToMesh("export.x");
		if(ch==VK_F4) editor_mode = false, InitLevel(level_id,INIT_EDITOR_LEAVE);
		if(ch==VK_F5) InitLevel(level_id,INIT_EDITOR_LOAD);
		if(ch==VK_F6)
		{
			chmap.Save(format("maps/%04d.bin",level_id).c_str());
			WriteConfig(format("maps/%04d.cfg",level_id).c_str());
		}
		if(ch==VK_F7 && level_id>   0) InitLevel(level_id-1,INIT_EDITOR_LOAD);
		if(ch==VK_F8 && level_id<9999) InitLevel(level_id+1,INIT_EDITOR_LOAD);
		if(ch==VK_OEM_3) radius = shift ? 0.005f : 0.001f;
		if(ch=='1') radius = shift ? 0.014f : 0.010f;
		if(ch=='2') radius = shift ? 0.025f : 0.018f;
		if(ch=='3') radius = shift ? 0.038f : 0.030f;
		if(ch=='4') radius = shift ? 0.058f : 0.045f;
		if(ch=='5') radius = shift ? 0.090f : 0.070f;
		if(ch=='U') chmap.Quantize();
		if( cameras.size()>0 )
		{
			if( ch=='B' ) locations[0] = -cameras[0].dir_front, InitLevel(level_id, INIT_BALL_RESET);
			if( ch=='N' ) locations[1] = -cameras[0].dir_front, InitLevel(level_id, INIT_BALL_RESET);
			if( ch=='C' ) locations[2] = cameras[0].dir_front, locations[3] = cameras[0].dir_up;
			if( ch=='X' ) cameras[0].dir_front = locations[2], cameras[0].dir_up = locations[3];
			if( ch=='T' ) locations[4] = -cameras[0].dir_front;
			if( ch=='Y' ) locations[5] = -cameras[0].dir_front;
		}
		//if(ch=='G') ball_types[0]++, ball_types[0]%=NUM_BALL_TYPES, InitLevel(level_id,INIT_BALL_RESET);
		//if(ch=='H') ball_types[1]++, ball_types[1]%=NUM_BALL_TYPES, InitLevel(level_id,INIT_BALL_RESET);
		if(ch=='J') mat[0].LoadNext(material_list);
		if(ch=='K') mat[1].LoadNext(material_list);
		if(ch=='L') mat[2].LoadNext(material_list);
		if(ch>='A' && ch<='Z') ProcessConfig(format("maps/macro_%c.cfg",ch).c_str(),0);
	}

	Dev.DPrintF("EDITOR mode, level %04d (press F1 for help)", level_id);
	
	if(editor_help)
		for(int i=0;editor_help_text[i];i++)
			Dev.DPrintF(editor_help_text[i]);

	chmap.UpdateTexture();
}

void Engine::RenderView(EngineCamera &cam)
{
	cam.viewport.MinZ = 0;
	cam.viewport.MaxZ = 1;
	Dev->SetViewport(&cam.viewport);

	// Setup technique params
	{
		vec3 front = cam.dir_front.get_normalized();
		vec3 right = cam.dir_up.cross(cam.dir_front).get_normalized();
		vec3 up = (cam.dir_up - front*front.dot(cam.dir_up)).get_normalized();

		fx.SetFloat2("vppos", vec2(cam.viewport.X, cam.viewport.Y));
		fx.SetFloat2("vpsize", vec2(cam.viewport.Width, cam.viewport.Height));
		fx.SetFloat("vzoom", cam.zoom);
		fx.SetFloat3("dir_front", front);
		fx.SetFloat3("dir_right", right);
		fx.SetFloat3("dir_up", up);
	}

	// Render traced objects
	{
		fx.StartTechnique(m_ExtendedShaderFeaturesEnabled ? "tech_full" : "tech_limited");
		fx->SetTexture("t_cube_c", chmap.GetMainTexture());
		fx->SetTexture("t_cube_h", chmap.GetHeightTexture());

		for( int i=0; i<5; i++ )
			ball[i].pos = vec3(0, 0, 0.01f);

		for( int i=0; i<(int)objects.size(); i++ )
		{
			EngineObject *ob = objects[i].get();
			if( !ob )
			{
				objects[i--] = objects.back();
				objects.pop_back();
				continue;
			}

			if( ob->render_mode==3 && ob->render_attach>=0 && ob->render_attach<4 )
				ball[ob->render_attach].pos = ob->pos;
		}

		float glow = pow(sin(Dev.GetElapsedTime()*2*3.1411593)*0.5f+0.5f,3)*20+1;

		fx.SetFloat3("ball1_pos", ball[0].pos);
		fx.SetFloat3("ball2_pos", ball[1].pos);
		fx.SetFloat3("ball3_pos", ball[2].pos);
		fx.SetFloat3("ball4_pos", ball[3].pos);
		fx.SetFloat3("ball5_pos", ball[4].pos);
		fx.SetFloat3("ball1_color", ball[0].color*glow);
		fx.SetFloat3("ball2_color", ball[1].color*glow);
		fx.SetFloat3("ball3_color", ball[2].color*glow);
		fx.SetFloat3("ball4_color", ball[3].color*glow);
		fx.SetFloat3("ball5_color", ball[4].color*glow);
		fx.SetFloat3("exit1_pos", locations[4]);
		fx.SetFloat3("exit2_pos", locations[5]);
		fx.SetFloat3("exit_status", exit_status);

		float pp[8] ={
			mat[2].m_mist_min, mat[2].m_mist_max, mat[2].m_heat*cam.viewport.Height/768.f, mat[2].m_mist_density,
			mat[2].m_mistcolor.x, mat[2].m_mistcolor.y, mat[2].m_mistcolor.z, 0
		};

		float epp[8] ={
			0.70f, 1.00f, 4.0f, 0.5f,
			1-pow(1.5f,-mat[2].m_exitcolor.x*14),
			1-pow(1.5f,-mat[2].m_exitcolor.y*14),
			1-pow(1.5f,-mat[2].m_exitcolor.z*14),
			0
		};

		for( int i=0; i<8; i++ )
			pp[i] += (epp[i]-pp[i])*exit_status.z;

		fx->SetFloatArray("m_mist_params", pp, 4);
		fx->SetFloatArray("m_mist_color", pp+4, 4);
		fx->SetFloatArray("exit_color", &(mat[2].m_exitcolor.x), 3);
		fx->SetFloatArray("exit_color2", epp+4, 3);

		mat[0].Set(fx, "0");
		mat[1].Set(fx, "1");
		mat[2].Set(fx, "2");

		Dev.SetRState(0);

		while( fx.StartPass() )
			Dev.DrawScreenQuad((float)cam.viewport.X, (float)cam.viewport.Y, (float)(cam.viewport.X+cam.viewport.Width), (float)(cam.viewport.Y+cam.viewport.Height), 1.f);
	}

	// Render quad objects
	{
		vb.clear();
		//Vertex v;
		//v.pos = vec3(-1, -1, 0);	v.texCoord=TODO;	vb.push_back(v);
		//v.pos = vec3( 1, -1, 0);	v.texCoord=TODO;	vb.push_back(v);
		//v.pos = vec3( 1,  1, 0);	v.texCoord=TODO;	vb.push_back(v);
		//v.pos = vec3(-1,  1, 0);	v.texCoord=TODO;	vb.push_back(v);
		for( int i=0; i<(int)objects.size(); i++ )
		{
			EngineObject *ob = objects[i].get();
			if( !ob ) continue;

			if( ob->render_mode == 1 )
			{
				const size_t frameIndex = ob->GetCurrentFrameIndex();
				TexturePackerDocument::Rect srcTextureRect;
				if( frameIndex >= 0 && frameIndex < (int)m_TexturePackerDoc.m_Frames.size() )
					srcTextureRect = m_TexturePackerDoc.m_Frames[frameIndex].frame;
				else
					srcTextureRect ={ 0, 0, m_TexturePackerDoc.m_Size.w, m_TexturePackerDoc.m_Size.h };

				float tx1 = (float)srcTextureRect.x * m_TexturePackerDoc.m_SizeX_inv;
				float ty2 = (float)srcTextureRect.y * m_TexturePackerDoc.m_SizeY_inv;
				float tx2 = (float)(srcTextureRect.x + srcTextureRect.w) * m_TexturePackerDoc.m_SizeX_inv;
				float ty1 = (float)(srcTextureRect.y + srcTextureRect.h) * m_TexturePackerDoc.m_SizeY_inv;

				// Small/Big = 0.025 / 0.05
				int layers = ob->render_layers;
				if( layers<1 ) layers = 1;
				if( layers>100 ) layers = 100;
				float thick = layers>1 ? ob->render_thickness / (layers-1) : 0;	// spacing
				for(int l=0;l<layers;l++)
				{
					float ss = 1.f;
					float lz = thick*l;
					Vertex v;
					vec3 pos = ob->pos + ob->pos.get_normalized()*lz;
					vec3 front = ob->front * ob->size.y * ss;
					vec3 side = ob->front.cross(ob->pos).get_normalized() * ob->size.x * ss;
					v.pos = pos - side + front;	v.texCoord = vec2(tx1, ty2);	vb.push_back(v);
					v.pos = pos + side + front;	v.texCoord = vec2(tx2, ty2);	vb.push_back(v);
					v.pos = pos + side - front;	v.texCoord = vec2(tx2, ty1);	vb.push_back(v);
					v.pos = pos - side - front;	v.texCoord = vec2(tx1, ty1);	vb.push_back(v);
				}
			}
		}

		if( vb.size()>0 )
		{
			fx.StartTechnique("object");
			Dev->SetFVF(Vertex::FVF);
            SetupObjectTexture();
			while( fx.StartPass() )
				Dev->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, vb.size(), vb.size()/4*2, Dev.GetQuadsIndices(vb.size()/4), D3DFMT_INDEX32, &vb[0], sizeof(Vertex));
		}
	}

	// Render mesh objects
	for( int i=0; i<(int)objects.size(); i++ )
	{
		EngineObject *ob = objects[i].get();
		if( !ob || ob->render_mode!=2 || !ob->render_mesh || ob->render_mesh->vb.size()<=0 || ob->render_mesh->ib.size()<=0 ) continue;
		EngineMesh *m = ob->render_mesh;
		Dev->SetFVF(EngineMesh::Vertex::FVF);
		fx.SetFloat4("obj_pos", *(vec4*)&ob->pos);
		fx.SetFloat3("obj_front", ob->front);

		fx.StartTechnique("object3d");
		while( fx.StartPass() )
			Dev->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, m->vb.size(), m->ib.size()/3, &m->ib[0], D3DFMT_INDEX32, &m->vb[0], sizeof(EngineMesh::Vertex));
	}
}

void Engine::RenderAll()
{
	g_time += Dev.GetTimeDelta();

	chmap.UpdateTexture();

	// Setup global rendering params
	int sx, sy;
	Dev.GetScreenSize(sx, sy);
	fx->SetFloat("time", g_time);
	fx->SetFloatArray("ssize", (float*)&vec2((float)sx, (float)sy), 2);

	// Target: buffers
	Dev.SetRenderTarget(0, &main_rt);
	Dev.SetRenderTarget(1, &heat_rt);
	Dev->Clear(0, NULL, D3DCLEAR_TARGET, 0x000000, 1.f, 0);

	// Render views
	{
		int cam_count = (int)cameras.size();	// including empty slots
		int cams_per_row = 1;
		if( cam_count>=2 ) cams_per_row = 2;
		if( cam_count>=5 ) cams_per_row = 3;
		if( cam_count>=10 ) cams_per_row = 4;
		if( cam_count>=17 ) cams_per_row = 5;
		if( cam_count>=26 ) cams_per_row = 6;

		int cam_rows = (cam_count+cams_per_row-1)/cams_per_row;
		float cam_zoom = 1.f;
		float cam_vp_x = sx/cams_per_row;
		float cam_vp_y = sy/cam_rows;
		float cam_numsq = cams_per_row * cam_rows;
		if( !editor_mode )
			cam_zoom = 1.2f * sqrt((3+cam_numsq)/4) * pow(cam_vp_x/cam_vp_y, 0.3f);

		for( int i=0; i<(int)cameras.size(); i++ )
		{
			EngineCamera &cam = cameras[i];
			int cam_row_y = i/cams_per_row;
			int cam_row_x = i - cams_per_row*cam_row_y;
			int cams_in_row = cams_per_row;
			if( cam_count - cams_per_row*cam_row_y < cams_in_row )
				cams_in_row = cam_count - cams_per_row*cam_row_y;

			cam.viewport.X = sx*cam_row_x/cams_in_row;
			cam.viewport.Y = sy*cam_row_y/cam_rows;
			cam.viewport.Width = sx*(cam_row_x+1)/cams_in_row - cam.viewport.X;
			cam.viewport.Height = sy*(cam_row_y+1)/cam_rows - cam.viewport.Y;
			cam.zoom = cam_zoom;

			if( cam.active )
				RenderView(cam);
		}
	}

	// Switch to main screen
	Dev.SetRenderTarget(0, NULL);
	Dev.SetRenderTarget(1, NULL);

	// Render heat haze
	fx.StartTechnique("heat");
	fx->SetTexture("t_screen", main_rt.GetTexture());
	fx->SetTexture("t_heat", heat_rt.GetTexture());
	fx->SetTexture("t_random", t_random.GetTexture());
	while( fx.StartPass() )
		Dev.DrawScreenQuad(0, 0, (float)sx, (float)sy);

	// TBD: postprocessing
	//PostProcess();
}

EngineMesh *Engine::GetMesh(const char *path)
{
	auto p = e_meshes.find(path);
	if( p!=e_meshes.end() )
		return p->second;
	
	EngineMesh *mesh = new EngineMesh();
	mesh->LoadPly(path);
	e_meshes[path] = mesh;
	return mesh;
}


void Engine::SetLocation(int loc, float x, float y, float z)
{
	if( loc<0 || loc>=MAX_LOCATIONS )
		return;
	locations[loc].x = x;
	locations[loc].y = y;
	locations[loc].z = z;
}

void Engine::UnlinkAllObjects()
{
	objects.clear();
}

void Engine::LinkObject(EngineObject *ob)
{
	for( int i=0; i<(int)objects.size(); i++ )
		if( objects[i]==ob )
			return;
	objects.push_back(ob);
}

void Engine::UnlinkObject(EngineObject *ob)
{
	for( int i=0; i<(int)objects.size(); i++ )
		if( objects[i]==ob )
		{
			objects[i] = objects.back();
			objects.pop_back();
			return;
		}
}

const Animation* Engine::GetAnimationByName(const char* name) const
{
    if(strcmp(name, "BugWalking") == 0)
        return &m_AnimBugWalking;
    if(strcmp(name, "GuyWalking") == 0)
        return &m_AnimGuyWalking;
	if (strcmp(name, "LaserBugWalking") == 0)
		return &m_AnimLaserBugWalking;
	if (strcmp(name, "LaserBugShoot") == 0)
		return &m_AnimLaserBugShoot;
    return nullptr;
}

void Engine::SetupObjectTexture()
{
    fx->SetTexture("t_object", g_TexturePackerTexture.GetTexture());

    Dev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
    Dev->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    //Dev->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_TFACTOR);
    Dev->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
    Dev->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
    //Dev->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_TFACTOR);

    //Dev->SetRenderState(D3DRS_TEXTUREFACTOR, 0xFFFF8080);
}

void Engine::InitAnimations()
{
    const char* bugWalkingFrameNames[] = {
        "bite_0000.png",
        "bite_0001.png",
        "bite_0002.png",
        "bite_0003.png",
        "bite_0004.png",
        "bite_0003.png",
        "bite_0002.png",
        "bite_0001.png",
    };
    m_AnimBugWalking.Set(
        bugWalkingFrameNames, _countof(bugWalkingFrameNames), 0.2f);

    const char* guyWalkingFrameNames[] = {
        "guy_0000.png",
        "guy_0001.png",
    };
    m_AnimGuyWalking.Set(
        guyWalkingFrameNames, _countof(guyWalkingFrameNames), 0.2f);

	const char* laserBugWalkingFrameNames[] = {
		"laser0000.png",
		"laser0005.png",
		"laser0006.png",
		"laser0007.png",
		"laser0008.png",
	};
	m_AnimLaserBugWalking.Set(
		laserBugWalkingFrameNames, _countof(laserBugWalkingFrameNames), 0.2f);

	const char* laserBugShootFrameNames[] = {
		"laser0000.png",
		"laser0001.png",
		"laser0002.png",
		"laser0003.png",
		"laser0004.png",
	};
	m_AnimLaserBugShoot.Set(
		laserBugShootFrameNames, _countof(laserBugShootFrameNames), 0.2f);
}

void Engine::InitSoundEffects()
{
    m_SoundEffects[SoundEffect_Laser].sound = new Sound::Sound(
        "data/lasergun.wav", 1.f, false);
    m_SoundEffects[SoundEffect_FootSteps].sound = new Sound::Sound(
        "data/footsteps.ogg", 1.f, true);
    m_SoundEffects[SoundEffect_GemPickUp].sound = new Sound::Sound(
        "data/gemPickUP.ogg", 1.f, false);
	m_SoundEffects[SoundEffect_LaserBigBug].sound = new Sound::Sound(
		"data/laser_big_bug.ogg", 1.f, false);
	m_SoundEffects[SoundEffect_LaserSmallBug].sound = new Sound::Sound(
		"data/laser_small_bug.ogg", 1.f, false);
}

void Engine::SetCameraCount(int count)
{
	int cam_i = cameras.size();

	cameras.resize(count);
	for( ; cam_i<(int)cameras.size(); cam_i++ )
	{
		cameras[cam_i].dir_front = locations[2].get_normalized();
		cameras[cam_i].dir_up = locations[3].get_normalized();
	}
}

void Engine::AttachCamera(int id, EngineObject *obj)
{
	if( id>=(int)cameras.size() )
		SetCameraCount(id+1);
	cameras[id].attach_obj = obj;
	cameras[id].active = true;
}

void Engine::RemoveCamera(int id)
{
	if( uint32_t(id)<cameras.size() )
	{
		cameras[id].attach_obj = NULL;
		cameras[id].active = false;
	}
}

float Engine::GetCamFrontX(int id) { return (uint32_t(id)<cameras.size()) ? cameras[id].dir_front.x : 0; }
float Engine::GetCamFrontY(int id) { return (uint32_t(id)<cameras.size()) ? cameras[id].dir_front.y : 0; }
float Engine::GetCamFrontZ(int id) { return (uint32_t(id)<cameras.size()) ? cameras[id].dir_front.z : 0; }
float Engine::GetCamUpX(int id) { return (uint32_t(id)<cameras.size()) ? cameras[id].dir_up.x : 0; }
float Engine::GetCamUpY(int id) { return (uint32_t(id)<cameras.size()) ? cameras[id].dir_up.y : 0; }
float Engine::GetCamUpZ(int id) { return (uint32_t(id)<cameras.size()) ? cameras[id].dir_up.z : 0; }

size_t Engine::SoundEffectNameToIndex(const char* name)
{
    if(_stricmp(name, "Laser") == 0)
        return SoundEffect_Laser;
    if(_stricmp(name, "FootSteps") == 0)
        return SoundEffect_FootSteps;
    if(_stricmp(name, "GemPickUp") == 0)
        return SoundEffect_GemPickUp;
	if (_stricmp(name, "LaserBigBug") == 0)
		return SoundEffect_LaserBigBug;
	if (_stricmp(name, "LaserSmallBug") == 0)
		return SoundEffect_LaserSmallBug;
    return SoundEffect_Count;
}

void Engine::PlaySoundEffect(const char* name, float volume)
{
    size_t index = SoundEffectNameToIndex(name);
    if(index < SoundEffect_Count)
    {
        SoundInfo& info = m_SoundEffects[index];
        assert(!info.sound->IsLoop());
        info.sound->Play(volume);
    }
    else
    {
        assert(0);
    }
}

void Engine::StartSoundEffectLoop(const char* name, float volume)
{
    size_t index = SoundEffectNameToIndex(name);
    if(index < SoundEffect_Count)
    {
        SoundInfo& info = m_SoundEffects[index];
        assert(info.sound->IsLoop());
        if(info.loopedChannel != nullptr)
        {
            info.sound->Stop(info.loopedChannel);
        }
        info.loopedChannel = info.sound->Play(volume);
    }
    else
    {
        assert(0);
    }
}

void Engine::StopSoundEffectLoop(const char* name)
{
    size_t index = SoundEffectNameToIndex(name);
    if(index < SoundEffect_Count)
    {
        SoundInfo& info = m_SoundEffects[index];
        assert(info.sound->IsLoop());
        if(info.loopedChannel != nullptr)
        {
            info.sound->Stop(info.loopedChannel);
            info.loopedChannel = nullptr;
        }
    }
    else
    {
        assert(0);
    }
}

void Engine::RunProgram(
    const char* filePath,
    const char* parameters)
{
    string filePathCopy = filePath;
    string parametersCopy = parameters;
    HWND hwnd = Dev.GetHWnd();
    std::thread thread([hwnd, filePathCopy, parametersCopy]() {
        ShellExecute(
            hwnd,
            NULL,
            filePathCopy.c_str(),
            parametersCopy.c_str(),
            NULL,
            SW_SHOWDEFAULT);
    });
    thread.detach();
}

bool Engine::GetKeyState(int key)
{
    return Dev.GetKeyState(key);
}

bool Engine::GetKeyStroke(int key)
{
    return Dev.GetKeyStroke(key);
}

bool Engine::FindClosestFaction(float x, float y, float z, float radius, int faction)
{
	vec3 pos = vec3(x,y,z);
	EngineObject *found = NULL;

	radius *= radius;
	for( int i=0; i<(int)objects.size(); i++ )
	{
		EngineObject *ob = objects[i].get();
		if( !ob || !(ob->hit_faction&faction) ) continue;
		if( (ob->pos - pos).length2() < radius )
		{
			radius = (ob->pos - pos).length2();
			found = ob;
		}
	}

	if( found )
		Sqrat::RootTable(sq()).SetValue("other", found);

	return found != NULL;
}
