
#pragma once

#include "TexturePackerSupport.h"

namespace Sound {
    class Sound;
}
namespace FMOD
{
    class Channel;
}

class EngineObject;


extern const char* const TEXTURE_ATLAS_TEXTURE_FILE_PATH;


class Ball {
public:
	vec3	pos;
	vec3	color;

	void Init(vec3 &_pos)
	{
		pos = _pos;
		color = vec3(0.3f, 1.0f, 1.3f);
	}
};


#define MAX_LOCATIONS			6

#define INIT_CAMERA				1
#define INIT_MAP				2
#define INIT_MATERIALS			4
#define INIT_CONFIG				8
#define INIT_MESSAGES			16
#define INIT_RESET_COMPLETED	32
#define INIT_ALL				(INIT_CAMERA|INIT_MAP|INIT_MATERIALS|INIT_CONFIG|INIT_MESSAGES)
#define INIT_EDITOR_LOAD		(INIT_MAP|INIT_CONFIG|INIT_RESET_COMPLETED|INIT_MATERIALS)
#define INIT_BALL_RESET			0
#define INIT_EDITOR_ENTER		INIT_RESET_COMPLETED
#define INIT_EDITOR_LEAVE		0
#define INIT_LEVEL_RESTART		(INIT_CAMERA|INIT_RESET_COMPLETED)

class EngineCamera {
public:
	vec3							dir_front;
	vec3							dir_up;
	vec2							mouse_smooth = vec2(0, 0);
	D3DVIEWPORT9					viewport;
	float							zoom;
	base::weak_ptr<EngineObject>	attach_obj;
	bool							active = false;

	void RunCamera(float delta);
};

enum SoundEffect
{
    SoundEffect_Laser,
    SoundEffect_FootSteps,
    SoundEffect_GemPickUp,
	SoundEffect_LaserBigBug,
	SoundEffect_LaserSmallBug,
    SoundEffect_Count
};

class Engine {
public:
    TexturePackerDocument m_TexturePackerDoc;
	CubeHeightmap							chmap;
	Material								mat[3];
	Ball									ball[2];
	int										ball_types[2];
	vec3									locations[MAX_LOCATIONS];
	int										level_id;

	vector<EngineCamera>					cameras;

	bool									editor_mode;
	vector<string>							material_list;
	vec3									exit_status;
	vector<GameUI::Message>					messages;
	vector<string>							raw_cmd;
	int										completed;
	vector<base::weak_ptr<EngineObject> >	objects;


	Engine();
    void Init();
	void InitLevel(int level, int mode);
	void ProcessConfig(const char *path, int mode);
	void WriteConfig(const char *path);
	void Physics(float delta);

	void RunGame(float delta);
	void RunEditor(float delta);

	void Run(float delta)
	{
		if( editor_mode )
			RunEditor(delta);
		else
			RunGame(delta);
	}

	void RenderView(EngineCamera &cam);
	void RenderAll();

	EngineMesh *GetMesh(const char *path);


	// API functions
    float GetElapsedTime() { return (float)Dev.GetElapsedTime(); }
    float GetTimeDelta() { return Dev.GetTimeDelta(); }				// Time delta in seconds
	void SetLocation(int loc, float x, float y, float z);			// Set location vector (does various things depending on index)

	void UnlinkAllObjects();
	void LinkObject(EngineObject *ob);
	void UnlinkObject(EngineObject *ob);
    const Animation* GetAnimationByName(const char* name) const;

	void SetCameraCount(int count);
	void AttachCamera(int id, EngineObject *obj);
	void RemoveCamera(int id);
	float GetCamFrontX(int id);
	float GetCamFrontY(int id);
	float GetCamFrontZ(int id);
	float GetCamUpX(int id);
	float GetCamUpY(int id);
	float GetCamUpZ(int id);

    static size_t SoundEffectNameToIndex(const char* name);
        
    void PlaySoundEffect(const char* name, float volume);
    void StartSoundEffectLoop(const char* name, float volume);
    void StopSoundEffectLoop(const char* name);
    void RunProgram(
        const char* filePath,
        const char* parameters);
    bool GetKeyState(int key);
    bool GetKeyStroke(int key);

	bool FindClosestFaction(float x, float y, float z, float radius, int faction);


protected:
	// internals
	struct Vertex {
		enum { FVF = D3DFVF_XYZ | D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE2(0) };
		vec3	pos;
        vec2    texCoord;
	};
	vector<Vertex>	vb;
    Animation m_AnimBugWalking;
    Animation m_AnimGuyWalking;
	Animation m_AnimLaserBugWalking;
	Animation m_AnimLaserBugShoot;
	
	map<string,EngineMesh*>	e_meshes;			// not cleaned up on exit ;)

    bool m_ExtendedShaderFeaturesEnabled;

    struct SoundInfo
    {
        Sound::Sound* sound = nullptr;
        FMOD::Channel* loopedChannel = nullptr;
    } m_SoundEffects[SoundEffect_Count];

    void SetupObjectTexture();
    void InitAnimations();
    void InitSoundEffects();
};


extern Engine game;
