#include "pch.h"
#include "common.h"


void EngineObject::SetPosition(float x, float y, float z)
{
	pos.x = x;
	pos.y = y;
	pos.z = z;
}

void EngineObject::SetFront(float x, float y, float z)
{
	front.x = x;
	front.y = y;
	front.z = z;
	front.normalize();
}

void EngineObject::SetVelocity(float x, float y, float z)
{
	if( physics_gravity>0 )
	{
		vec3 vv = vec3(x, y, z);
		vec3 pn = pos.get_normalized();
		vv -= pn*pn.dot(vv);			// remove vertical component from new velocity
		vel = vv + pn*pn.dot(vel);		// keep old vertical component
	}
	else
	{
		vel.x = x;
		vel.y = y;
		vel.z = z;
	}
}

void EngineObject::SetVelocityForce(float x, float y, float z)
{
	vel.x = x;
	vel.y = y;
	vel.z = z;
}


void EngineObject::SetSize(float x, float y)
{
	size.x = x;
	size.y = y;
}

void EngineObject::SetStaticFrame(const char* name)
{
    m_StaticFrameIndex = game.m_TexturePackerDoc.FrameFileNameToIndex(name);
    m_Animation = nullptr;
}

void EngineObject::SetAnimation(const char* name)
{
    const Animation* const anim = game.GetAnimationByName(name);
    if(anim)
    {
        m_StaticFrameIndex = SIZE_MAX;
        m_Animation = anim;
        m_AnimStartTime = (float)Dev.GetElapsedTime();
    }
    else
    {
        assert(0);
    }
}

void EngineObject::SetMesh(const char *name)
{
	render_mode = 2;
	render_mesh = game.GetMesh(format("data/%s.ply", name).c_str());
}


void EngineObject::TickPhysics(float delta)
{
	if( physics_gravity>0 ) vel -= pos.get_normalized()*physics_gravity*delta;
	pos += vel*delta;

	if( physics_mode )
	{
		for( int i=0; i<3; i++ )
		{
			vec3 p;
			if( !game.chmap.FindNearestPoint(pos, physics_radius, p, -physics_sink) )		// size.x is "ball" radius
				break;
			vec3 n = (pos - p).get_normalized();
			if( n.dot(vel)<0 )
			{
				vel -= n*(n.dot(vel)*1.0f);

				if( p.length() + physics_radius*0.5f >= pos.length() )
				{
					if( physics_mode==2 )
						game.UnlinkObject(this);
					wall_collision++;
				}
			}
			pos = p + n*(physics_radius*1.01f);
		}
	}

	if( level_max )
	{
		float plen = pos.length();

		if( plen > 0 )
		{
			vec3 pn = pos/plen;

			front = (front - pn*pn.dot(front)).get_normalized();

			if( plen<level_min ) pos = pn*level_min;
			if( plen>level_max ) pos = pn*level_max;
		}
	}

	if( hit_colmask )
	{
		float r2 = physics_radius * physics_radius;
		float min_dist = 0;
		EngineObject *col_other = NULL;

		for( int i=0; i<(int)game.objects.size(); i++ )
		{
			EngineObject *other = game.objects[i].get();
			if( !other || !(other->hit_faction & hit_colmask) ) continue;

			float dist = (other->pos - pos).length2() - r2 - (other->physics_radius * other->physics_radius);
			if( dist < min_dist )
			{
				dist = min_dist;
				col_other = other;
			}
		}

		if( col_other )
		{
			Sqrat::RootTable(sq()).SetValue("self", this);
			Sqrat::RootTable(sq()).SetValue("other", col_other);
			sq.Run("on_object_hit");
		}
	}
}

size_t EngineObject::GetCurrentFrameIndex()
{
    if(m_Animation)
    {
        return m_Animation->GetCurrentFrameIndex((float)Dev.GetElapsedTime() - m_AnimStartTime);
    }
    else
    {
        return m_StaticFrameIndex;
    }
}
